from . import AbstractProject
from . import AbstractProjectItem
from . import DictItemInterface
from . import StringFloatItemInterface
from . import StringIntItemInterface
from . import StringItemInterface
from . import StringListItemInterface


class AbstractSetting(AbstractProject):
    pass


class AbstractSettingItem(AbstractProjectItem):
    def __init__(self, parent: AbstractSetting, default_value=None):
        super().__init__(parent)

        if default_value is not None:
            if isinstance(default_value, list):
                self.self_storage = default_value
            else:
                self.self_storage = str(default_value)

    def get_value(self):
        if self.parent is not None:
            ret = self.parent.get_property(self.name)
            if ret is not None:
                return ret
        return self.self_storage


class StringSettingItem(AbstractSettingItem, StringItemInterface):
    pass


class StringListSettingItem(AbstractSettingItem, StringListItemInterface):
    pass


class IntSettingItem(AbstractSettingItem, StringIntItemInterface):
    pass


class FloatSettingItem(AbstractSettingItem, StringFloatItemInterface):
    pass


class DictSettingItem(AbstractSettingItem, DictItemInterface):
    pass



import traceback

import typing

from .sender import SignalSender


class ValueModel:
    @property
    def value(self):
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)

    def get_value(self):
        pass

    def set_value(self, value):
        pass

    def get_string(self) -> typing.Optional[str]:
        return self.get_value()


class AttachAbility(ValueModel):
    AttachFlag = '_attach_{}'

    def _create(self, creator, args: tuple = (),
                finished_with: typing.Callable[[], object] = None):
        name = self.AttachFlag.format(self.current_name)

        obj = getattr(self, name, None)
        if obj is None:
            obj = creator(*args)
            setattr(self, name, obj)
            if finished_with:
                finished_with()
        return obj

    def assign(self, value):
        name = self.AttachFlag.format(self.current_name)
        setattr(self, name, value)

    @property
    def current_name(self):
        return traceback.extract_stack(None, 3)[0][2]


class ValueInterface(ValueModel):
    pass


class ChangedInterface(AttachAbility):
    @property
    def changed(self) -> SignalSender:
        return self._create(SignalSender, finished_with=self.set_changed_connection)

    # noinspection PyUnresolvedReferences
    def check_change(self, *_):
        if self.value != self.changed.last_value:
            self.changed.emit(self.value)

    def emit_changed(self):
        self.changed.emit(self.value)

    def set_changed_connection(self):
        pass


class AbstractItem(ValueInterface, ChangedInterface):
    pass


class Item(AbstractItem):
    pass


class AbstractProperty(ValueInterface, ChangedInterface):
    def __init__(self, parent: AbstractItem):
        self.parent = parent

    def set_changed_connection(self):
        self.parent.changed.connect(self.check_change)

    def get_value(self):
        return self.parent.get_value()

    def set_value(self, value):
        self.parent.set_value(value)

    def connect(self, other: 'AbstractProperty'):
        self.changed.connect(other.set_value)
        other.changed.connect(self.set_value)
        other.value = self.value



import unittest

import prett6


class MyTestCase(unittest.TestCase):
    def test_dict_item_emit_change_condition(self):
        class ProjectDemo(prett6.AbstractProject):
            def __init__(self):
                self.d = prett6.DictProjectItem(self)

        p = ProjectDemo()
        self.emit_value = None

        @prett6.connect_with(p.d.changed)
        def copy_value_from_dict_item(data: dict):
            self.emit_value = data.copy()

        # emit changed if use `=` operator
        p.d.dict.value = {1: 2, 3: 4}
        self.assertEqual(self.emit_value, {1: 2, 3: 4})

        # can't emit change if modify value directly
        p.d.dict.value.pop(1)
        self.assertEqual(self.emit_value, {1: 2, 3: 4})
        self.assertEqual(p.d.dict.value, {3: 4})

        # can't emit changed if assign reference origin value to value
        p.d.dict.value = {1: 2, 3: 4}
        self.assertEqual(self.emit_value, {1: 2, 3: 4})
        d = p.d.dict.value  # reference value
        d.pop(1)
        p.d.dict.value = d
        self.assertEqual(self.emit_value, {1: 2, 3: 4})

        # emit changed if you copy origin value, change it and assign to value
        p.d.dict.value = {1: 2, 3: 4}
        d = p.d.dict.value.copy()  # reference value
        d.pop(1)
        p.d.dict.value = d
        self.assertEqual(self.emit_value, {3: 4})

        # emit change if using DictProjectItem method
        p.d.dict.value = {1: 2, 3: 4}
        p.d.dict.pop(1)
        self.assertEqual(self.emit_value, {3: 4})

    def test_list_item_emit_change_condition(self):
        class ProjectDemo(prett6.AbstractProject):
            def __init__(self):
                self.ls = prett6.ListProjectItem(self)

        p = ProjectDemo()
        self.emit_value = None

        @prett6.connect_with(p.ls.changed)
        def copy_value_from_dict_item(data: list):
            self.emit_value = data.copy()

        # emit changed if use `=` operator
        p.ls.list.value = [1, 2, 3, 4]
        self.assertEqual(self.emit_value, [1, 2, 3, 4])

        # can't emit change if modify value directly
        p.ls.list.value.pop(0)
        self.assertEqual(self.emit_value, [1, 2, 3, 4])
        self.assertEqual(p.ls.list.value, [2, 3, 4])

        # can't emit changed if assign reference origin value to value
        p.ls.list.value = [1, 2, 3, 4]
        self.assertEqual(self.emit_value, [1, 2, 3, 4])
        d = p.ls.list.value  # reference value
        d.pop(1)
        p.ls.list.value = d
        self.assertEqual(self.emit_value, [1, 2, 3, 4])

        # emit changed if you copy origin value, change it and assign to value
        p.ls.list.value = [1, 2, 3, 4]
        d = p.ls.list.value.copy()  # reference value
        d.pop(0)
        p.ls.list.value = d
        self.assertEqual(self.emit_value, [2, 3, 4])

        # emit change if using DictProjectItem method
        p.ls.list.value = [1, 2, 3, 4]
        p.ls.list.pop(0)
        self.assertEqual(self.emit_value, [2, 3, 4])


if __name__ == '__main__':
    unittest.main()



from PySide6.QtCore import QObject, Signal

__all__ = ('SignalSender', 'connect_with')


class SignalWrapper(QObject):
    signal = Signal(object)


class SignalSender:
    def __init__(self):
        self.signal = SignalWrapper()
        self.last_emit = None
        self.last_error = []

    def emit(self, *args, **kwargs):
        self.last_emit = args, kwargs
        self.last_error.clear()

        # 添加检查以避免访问已删除的信号实例
        try:
            self.signal.signal.emit((args, kwargs))
        except RuntimeError as e:
            if "was already deleted" in str(e):
                # 信号实例已被删除，记录警告并返回
                import logging
                logging.warning("Attempted to emit deleted signal: %s", str(e))
                return
            else:
                # 其他运行时错误，重新抛出
                raise

        if self.last_error:
            raise self.last_error[0]

    def connect(self, func, *args, **kwargs):
        last_error = self.last_error

        def slot_func(data, *_):
            try:

                if len(kwargs):
                    k = data[1].copy()
                    k.updata(kwargs)
                else:
                    k = data[1]

                return func(*(args + data[0]), **k)
            except BaseException as e:
                last_error.append(e)

        slot_func.__name__ = getattr(func, '__name__', '<Lambda>')

        self.signal.signal.connect(slot_func)

    def emit_if_changed(self, *args, **kwargs):
        emit_value = (args, kwargs)
        if self.last_emit != emit_value:
            self.emit(*args, **kwargs)

    @property
    def has_emitted(self):
        return self.last_emit is not None

    @property
    def last_value(self):
        if self.last_emit is None:
            return None
        else:
            args, kwargs = self.last_emit

            if len(kwargs) == 0:
                if len(args) == 0:
                    return None
                elif len(args) == 1:
                    return args[0]
                else:
                    return args
            else:
                return self.last_emit


def connect_with(signal: SignalSender, *args, **kwargs):
    def pack_func(func):
        signal.connect(func, *args, **kwargs)
        return func

    return pack_func



import typing

from . import AbstractItem
from . import AbstractProperty
from . import ValueModel


class StringValueModel(ValueModel):
    @property
    def value(self) -> str:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class StringProperty(AbstractProperty, StringValueModel):
    pass


class StringItemInterface(AbstractItem):
    @property
    def string(self) -> StringProperty:
        return self._create(StringProperty, args=(self,))


class StringItem(StringItemInterface):
    pass


class IntValueModel(ValueModel):
    @property
    def value(self) -> int:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(int(value))


class IntProperty(AbstractProperty, IntValueModel):
    pass


class IntItemInterface(AbstractItem):
    @property
    def int(self) -> IntProperty:
        return self._create(IntProperty, args=(self,))


class IntItem(IntItemInterface):
    pass


class FloatValueModel(ValueModel):
    @property
    def value(self) -> float:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(float(value))


class FloatProperty(AbstractProperty, FloatValueModel):
    pass


class FloatItemInterface(AbstractItem):
    @property
    def float(self) -> FloatProperty:
        return self._create(FloatProperty, args=(self,))


class FloatItem(FloatItemInterface):
    pass


class DictValueModel(ValueModel):
    @property
    def value(self) -> dict:
        if self.get_value() is None:
            return {}
        if isinstance(self.get_value(), dict):
            return self.get_value()
        else:
            return eval(self.get_value())

    @value.setter
    def value(self, value):
        self.set_value(value)


class DictProperty(AbstractProperty, DictValueModel):

    def __setitem__(self, key, value):
        self.value.__setitem__(key, value)
        self.parent.emit_changed()
        self.emit_changed()

    def __getitem__(self, item):
        return self.value.__getitem__(item)

    def __getattr__(self, attr_name):
        if not self.__dict__.__contains__(attr_name) and attr_name in dir(self.value):
            def wrapper(*args, **kwargs):
                previous_value = self.value.copy()
                return_value = getattr(self.value, attr_name)(*args, **kwargs)
                if previous_value != self.value:
                    # print(self)
                    self.parent.emit_changed()
                    self.emit_changed()
                return return_value

            return wrapper
        return self.__getattribute__(attr_name)


class DictItemInterface(AbstractItem):

    @property
    def dict(self) -> DictProperty:
        return self._create(DictProperty, args=(self,))


class DictItem(DictItemInterface):
    pass


class DictListValueModel(ValueModel):
    @property
    def value(self) -> typing.List[dict]:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)

    @property
    def count(self):
        return len(self.value)


class DictListProperty(AbstractProperty, DictListValueModel):
    pass


class DictListItemInterface(AbstractItem):
    @property
    def dict_list(self) -> DictListProperty:
        return self._create(DictListProperty, args=(self,))


class DictListItem(DictListItemInterface):
    pass


class ListValueModel(ValueModel):
    @property
    def value(self) -> list:
        if self.get_value() == '' or self.get_value() is None:
            return []
        if isinstance(self.get_value(), str):
            return eval(self.get_value())
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class ListProperty(AbstractProperty, ListValueModel):
    def __setitem__(self, key, value):
        self.value.__setitem__(key, value)
        self.parent.emit_changed()
        self.emit_changed()

    def __getitem__(self, item):
        return self.value.__getitem__(item)

    def __getattr__(self, attr_name):
        if not self.__dict__.__contains__(attr_name) and attr_name in dir(self.value):
            def wrapper(*args, **kwargs):
                previous_value = self.value.copy()
                return_value = getattr(self.value, attr_name)(*args, **kwargs)
                if previous_value != self.value:
                    self.parent.emit_changed()
                    self.emit_changed()
                return return_value

            return wrapper
        return self.__getattribute__(attr_name)

    pass


class ListItemInterface(AbstractItem):

    @property
    def list(self) -> ListProperty:
        return self._create(ListProperty, args=(self,))


class ListItem(ListItemInterface):
    pass


class StringListValueModel(ValueModel):
    @property
    def value(self) -> typing.List[str]:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)

    @property
    def count(self):
        return len(self.value)


class StringListProperty(AbstractProperty, StringListValueModel):
    pass


class StringListItemInterface(AbstractItem):
    @property
    def string_list(self) -> StringListProperty:
        return self._create(StringListProperty, args=(self,))


class StringListItem(StringListItemInterface):
    pass


class StringIntProperty(IntProperty):
    def get_value(self):
        return int(self.parent.get_value())

    def set_value(self, value):
        self.parent.set_value(str(value))


class StringIntItemInterface(StringItemInterface):
    @property
    def int(self) -> StringIntProperty:
        return self._create(StringIntProperty, args=(self,))


class StringFloatProperty(FloatProperty):
    def get_value(self):
        if self.parent.get_value() == '':
            return None
        return float(self.parent.get_value())

    def set_value(self, value):
        self.parent.set_value(str(value))


class StringFloatItemInterface(StringItemInterface):
    @property
    def float(self) -> StringFloatProperty:
        return self._create(StringFloatProperty, args=(self,))



import codecs
import enum
import json
import os
import time
import typing

from . import AbstractProperty
from . import ChangedInterface
from . import DictItemInterface
from . import DictValueModel
from . import ListItemInterface
from . import StringFloatItemInterface
from . import StringIntItemInterface
from . import StringItemInterface
from . import StringProperty
from . import ValueModel


class SaveInterface(ValueModel):
    def load(self, check_key: str, check_value: str):
        if not os.path.exists(self.filename):
            raise FileNotFoundError('Setting File [{}] Not Found'.format(self.filename))

        try:
            with codecs.open(self.filename, 'r', encoding='utf-8') as f:
                obj = json.load(f)
                assert isinstance(obj, dict)
                assert obj.get(check_key, None) == check_value
                self.value = obj
        except AssertionError as e:
            raise e
        except ValueError as e:
            raise e

    def save(self):
        with codecs.open(self.filename, 'w', encoding='utf-8') as f:
            json.dump(self.value, f, indent=2, ensure_ascii=False)

    @property
    def path(self) -> typing.Optional[str]:
        return None

    @property
    def name(self) -> typing.Optional[str]:
        return None

    @property
    def filename(self):
        return os.path.join(self.path, self.name)

    @property
    def is_existing(self):
        return os.path.exists(self.filename)


class AbstractProject(DictValueModel, SaveInterface, ChangedInterface):
    def get_value(self) -> dict:
        return self._create(dict)

    def set_value(self, value):
        self.value.clear()
        self.value.update(value)
        self.changed.emit(self.value)

    def get_property(self, name) -> object:
        return self.value.get(name)

    def set_property(self, name, value):
        self.value.update({name: value})
        # 添加检查以防止信号实例已被删除
        if hasattr(self, 'changed') and self.changed is not None:
            try:
                self.changed.emit(self.value)
            except RuntimeError as e:
                if "already deleted" in str(e):
                    # 信号实例已被删除，跳过发射
                    pass
                else:
                    # 重新抛出其他运行时错误
                    raise

    # children property
    @property
    def children(self) -> typing.List['AbstractProjectItem']:
        return self._create(list)

    # name property
    def __setattr__(self, key: str, value):
        if isinstance(value, AbstractProjectItem):
            value.name = key
        super().__setattr__(key, value)


class AbstractProjectItem(StringItemInterface):
    def __init__(self, parent: AbstractProject = None):
        self.parent = parent
        if parent is not None:
            self.parent.children.append(self)
            self.parent.changed.connect(self.check_change)

    def get_value(self):
        if self.parent is not None:
            return self.parent.get_property(self.name)
        else:
            return self.self_storage

    def set_value(self, value):
        if self.parent is not None:
            self.parent.set_property(self.name, value)
        else:
            self.self_storage = value
            self.check_change()

    @property
    def name(self) -> str:
        return self._create(str)

    @name.setter
    def name(self, value):
        self.assign(value)

    @property
    def self_storage(self):
        return self._create(lambda: None)

    @self_storage.setter
    def self_storage(self, value):
        self.assign(value)


class DictProjectItem(AbstractProjectItem, DictItemInterface):
    pass


class ListProjectItem(AbstractProjectItem, ListItemInterface):
    pass


class StringProjectItem(AbstractProjectItem):
    pass


class IntProjectItem(AbstractProjectItem, StringIntItemInterface):
    pass


class FloatProjectItem(AbstractProjectItem, StringFloatItemInterface):
    pass


class TimePointItem(StringProjectItem):

    def __init__(self, parent, t_format='%Y-%m-%d %H:%M:%S'):
        super().__init__(parent)
        self.t_format = t_format

    class TimePointProperty(StringProperty):
        def __init__(self, parent, t_format):
            super().__init__(parent)
            self.t_format = t_format

        def update(self):
            self.value = time.strftime(self.t_format)

    @property
    def time(self):
        return self._create(TimePointItem.TimePointProperty, args=(self, self.t_format))


class Enum(enum.Enum):
    @classmethod
    def get_values_list(cls) -> []:
        res = []
        for s in cls:
            res.append(s.value)
        return res

    @classmethod
    def get_key_by_value(cls, value: str, default_value=None):
        for s in cls:
            if s.value == value:
                return s
        else:
            return default_value

    def __eq__(self, s: str):
        return self.name == s


class EnumValueModel(ValueModel):
    @property
    def value(self):
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class EnumItem(AbstractProjectItem, StringItemInterface):
    def __init__(self, parent, e=None):
        super().__init__(parent)
        self.enum = e

    class EnumProperty(AbstractProperty, EnumValueModel):
        def __init__(self, parent, e: Enum):
            super().__init__(parent)
            self.enum = e

        def get_value(self):
            text = self.parent.get_string()
            return self.enum.get_key_by_value(text)

        def set_value(self, value: Enum):
            self.parent.set_value(value.value)

    @property
    def type(self):
        return self._create(self.EnumProperty, args=(self, self.enum))



from .base import AbstractItem, Item, AbstractProperty
from .base import ValueInterface, ChangedInterface
from .base import ValueModel, AttachAbility
from .multi_types import DictListValueModel, DictListProperty, DictListItemInterface, DictListItem
from .multi_types import DictValueModel, DictProperty, DictItemInterface, DictItem
from .multi_types import FloatValueModel, FloatProperty, FloatItemInterface, FloatItem
from .multi_types import IntValueModel, IntProperty, IntItemInterface, IntItem
from .multi_types import ListValueModel, ListProperty, ListItemInterface, ListItem
from .multi_types import StringFloatProperty, StringFloatItemInterface
from .multi_types import StringIntProperty, StringIntItemInterface
from .multi_types import StringListValueModel, StringListProperty, StringListItemInterface, StringListItem
from .multi_types import StringValueModel, StringProperty, StringItemInterface, StringItem
from .project import AbstractProject, AbstractProjectItem
from .project import Enum, EnumValueModel, EnumItem
from .project import StringProjectItem, IntProjectItem, FloatProjectItem, DictProjectItem, ListProjectItem
from .project import TimePointItem
from .sender import SignalSender, connect_with
from .setting import AbstractSetting, AbstractSettingItem
from .setting import StringSettingItem, StringListSettingItem, IntSettingItem, FloatSettingItem, DictSettingItem
from .widget_interface import IndexItem, WidgetIndexInterface
from .widget_interface import StringsItem, WidgetStringListInterface
from .widget_interface import WidgetDictItem, WidgetDictInterface
from .widget_interface import WidgetDictListInterface
from .widget_interface import WidgetStringItem, WidgetStringInterface



import json
import prett6
import unittest


class MyTestCase(unittest.TestCase):
    def test_project_storage_tree(self):
        class ItemDemo(prett6.IntProjectItem):
            pass

        class ProjectDemo(prett6.AbstractProject):
            def __init__(self):
                self.width = ItemDemo(self)
                self.height = ItemDemo(self)

        p = ProjectDemo()
        p.width.int.value = 16
        p.height.int.value = 20
        self.assertEqual(p.value, {'width': '16', 'height': '20'})
        json_string = json.dumps(p.value)

        p.value = {'height': 17}
        self.assertEqual(p.height.int.value, 17)
        self.assertEqual(p.width.value, None)

        p.value = json.loads(json_string)
        self.assertEqual(p.width.string.value, '16')
        self.assertEqual(p.height.string.value, '20')

    def test_value_changed(self):
        class ItemDemo(prett6.IntProjectItem):
            pass

        class ProjectDemo(prett6.AbstractProject):
            def __init__(self):
                self.width = ItemDemo(self)
                self.height = ItemDemo(self)

        p = ProjectDemo()
        times = []

        @prett6.connect_with(p.width.int.changed)
        def width_changed(value: int):
            times.append(len(times))

            if len(times) == 1:
                self.assertEqual(value, 16)
            elif len(times) == 2:
                self.assertEqual(value, 20)
            elif len(times) == 3:
                self.assertEqual(value, 0)
            else:
                pass

        p.width.string.value = '16'
        p.width.value = '20'
        p.width.int.value = 0

        self.assertEqual(len(times), 3)

    def test_setting_value(self):
        class SettingItemDemo(prett6.IntSettingItem):
            pass

        class SettingDemo(prett6.AbstractSetting):
            def __init__(self):
                self.margin = SettingItemDemo(self, 5)

        s = SettingDemo()
        self.assertEqual(s.margin.int.value, 5)
        self.assertEqual(s.value, {})

        s.margin.int.value = 20
        self.assertEqual(s.margin.int.value, 20)
        self.assertEqual(s.value, {'margin': '20'})


if __name__ == '__main__':
    unittest.main()



from . import AttachAbility

from . import StringItem
from . import StringIntItemInterface
from . import StringFloatItemInterface

from . import StringProperty
from . import StringIntProperty
from . import StringFloatProperty

from . import IntItem
from . import IntProperty

from . import StringListItem
from . import StringListProperty

from . import DictItem
from . import DictProperty
from . import DictListItem
from . import DictListProperty


class WidgetStringItem(StringItem, StringIntItemInterface, StringFloatItemInterface):
    pass


class WidgetStringInterface(AttachAbility):
    class StringItem(WidgetStringItem):
        pass

    @property
    def string_item(self) -> StringItem:
        return self._create(type(self).StringItem, args=(self,))

    @property
    def string(self) -> StringProperty:
        return self.string_item.string

    @property
    def int(self) -> StringIntProperty:
        return self.string_item.int

    @property
    def float(self) -> StringFloatProperty:
        return self.string_item.float


class IndexItem(IntItem):
    pass


class WidgetIndexInterface(AttachAbility):
    class IndexItem(IndexItem):
        pass

    @property
    def index_item(self) -> IndexItem:
        return self._create(type(self).IndexItem, args=(self,))

    @property
    def index(self) -> IntProperty:
        return self.index_item.int


class StringsItem(StringListItem):
    pass


class WidgetStringListInterface(AttachAbility):
    class StringsItem(StringsItem):
        pass

    @property
    def strings_item(self) -> StringsItem:
        return self._create(type(self).StringsItem, args=(self,))

    @property
    def string_list(self) -> StringListProperty:
        return self.strings_item.string_list


class WidgetDictListInterface(AttachAbility):
    class DictListItem(DictListItem):
        pass

    @property
    def dict_list_item(self) -> DictListItem:
        return self._create(type(self).DictListItem, args=(self,))

    @property
    def dict_list(self) -> DictListProperty:
        return self.dict_list_item.dict_list


class WidgetDictItem(DictItem):
    pass


class WidgetDictInterface(AttachAbility):
    class DictItem(DictItem):
        pass

    @property
    def dict_item(self) -> DictItem:
        return self._create(type(self).DictItem, args=(self,))

    @property
    def dict(self) -> DictProperty:
        return self.dict_item.dict
